// Code generated by mkconsts.go. DO NOT EDIT.

//go:build goexperiment.regabiwrappers
// +build goexperiment.regabiwrappers

package goexperiment

const RegabiWrappers = true
const RegabiWrappersInt = 1
