// Code generated by mkconsts.go. DO NOT EDIT.

//go:build goexperiment.heapminimum512kib
// +build goexperiment.heapminimum512kib

package goexperiment

const HeapMinimum512KiB = true
const HeapMinimum512KiBInt = 1
