// Copyright 2019 The Go Authors. All rights reserved.
// Use of this source code is governed by a BSD-style
// license that can be found in the LICENSE file.

package runtime

const (
	_RCTL_LOCAL_DENY = 0x2

	_RCTL_LOCAL_MAXIMAL = 0x80000000

	_RCTL_FIRST = 0x0
	_RCTL_NEXT  = 0x1
)
