// Copyright 2021 The Go Authors. All rights reserved.
// Use of this source code is governed by a BSD-style
// license that can be found in the LICENSE file.

//go:build unix || (js && wasm)

package main_test

import (
	"os"
	"syscall"
)

func quitSignal() os.Signal {
	return syscall.SIGQUIT
}
