// Copyright 2022 The Go Authors. All rights reserved.
// Use of this source code is governed by a BSD-style
// license that can be found in the LICENSE file.

package comment

import "testing"

// See https://golang.org/issue/52353
func Test52353(t *testing.T) {
	ident("𫕐ﯯ")
}
