// Code generated by mkconsts.go. DO NOT EDIT.

//go:build !goexperiment.preemptibleloops
// +build !goexperiment.preemptibleloops

package goexperiment

const PreemptibleLoops = false
const PreemptibleLoopsInt = 0
